#include <stdio.h>
#include <lapacke.h>
#include <pthread.h>

#include <stdint.h>
#include <stdbool.h>
#include <time.h>

#include <string.h>
#include <errno.h>

#include "config.h"

#define HISTOGRAM_SIZE              (MAX_N/STEP)
#define CLOCKS_PER_MILLIS           (CLOCKS_PER_SEC/1000)
#define LOGISTICS_INITIAL_CONDITION (-800.)

#define PROGRESS_BAR_TOTAL HISTOGRAM_SIZE
#include "progress-bar.h"

uint32_t histogram[HISTOGRAM_SIZE];

// .data is a pointer because it should be allocated in the heap
// (.data DOES NOT fit in the stack 🤡)
typedef struct {
  double  *data;
  double  *ref_data;
  int32_t ipiv[MAX_N];
  size_t  id;
} Thread;

typedef struct {
  Thread threads[N_THREADS];
} Benchmarker;

uint32_t thread_run_benchmark(Thread *thread, size_t n)
{
  int64_t total_time = 0;
  clock_t start, end;

  for (size_t i = 0; i < N_TESTS; i++) {
    // Reinitializes the values in the .data array: this avoids progressively
    // moving larger values to the beginning of the array, which would decrise
    // decrease the number of row interchanges required for computations
    memcpy(thread->data, thread->ref_data, sizeof(double)*n*n);

    start = clock();
    (void)LAPACKE_dgetrf(LAPACK_ROW_MAJOR,
                         (lapack_int)n, (lapack_int)n,
                         thread->data, (lapack_int)n, thread->ipiv);
    end = clock();
    total_time += (end - start)/CLOCKS_PER_MILLIS;
  }

  return (uint32_t)(total_time/N_TESTS);
}

void *thread_benchmark(void *arg)
{
  Thread *thread = (Thread*) arg;

  // Computations are distributed evenly across threads
  for (size_t n = 1 + thread->id*STEP; n < MAX_N; n += STEP*N_THREADS) {
    uint32_t duration = thread_run_benchmark(thread, n);

    histogram[(n - 1)/STEP] = duration;
    progress_bar_inc();
  }

  return NULL;
}

Benchmarker benchmarker_new(double *ref_data)
{
  Benchmarker bench = {0};

  // These arrays will live for the entire duration of the program,
  // so we just leak them 🤡
  double *data = malloc(sizeof(double)*MAX_N*MAX_N*N_THREADS);
  if (data == NULL) {
    fprintf(stderr, "ERROR: Buy more RAM!\n");
    exit(EXIT_FAILURE);
  }

  for (size_t i = 0; i < N_THREADS; i++) {
    bench.threads[i].data     = data + i*MAX_N*MAX_N;
    bench.threads[i].ref_data = ref_data;
    bench.threads[i].id       = i;
  }

  return bench;
}

void benchmarker_run(Benchmarker *bench)
{
  pthread_t handles[N_THREADS] = {0};
  for (size_t i = 0; i < N_THREADS; i++) {
    if (pthread_create(&handles[i], NULL,
                       thread_benchmark,
                       &bench->threads[i]) != 0) {
      fprintf(stderr, "ERROR: Failed to spawn thread %lu!\n", i);
      exit(EXIT_FAILURE);
    }
  }

  for (size_t i = 0; i < N_THREADS; i++) {
    pthread_join(handles[i], NULL);
  }
}

int main(void)
{
  const char *output_path = "histogram.bin";
  FILE *output = fopen(output_path, "w");

  if (output == NULL) {
    fprintf(stderr,
            "ERROR: Coundn't open output file \"%s\": %s",
            output_path, strerror(errno));
    return EXIT_FAILURE;
  }

  // ========================================================================
  printf("INFO: Initializing random input data... ");

  double *ref_data = malloc(sizeof(double)*MAX_N*MAX_N);
  if (ref_data == NULL) {
    fprintf(stderr, "ERROR: Buy more RAM!\n");
    exit(EXIT_FAILURE);
  }

  // Pseudorandom data given by the logistic map
  double acc = LOGISTICS_INITIAL_CONDITION;
  for (size_t i = 0; i < MAX_N*MAX_N; i++) {
    ref_data[i] = acc;
    acc = 1000. - acc*acc/500.;
  }

  printf("done!\n");

  // ========================================================================
  printf("INFO: Benchmarking the dgetrf function on %ux%u matrices... (using %u threads)\n",
         MAX_N, MAX_N, N_THREADS);

  Benchmarker bench = benchmarker_new(ref_data);
  benchmarker_run(&bench);

  // ========================================================================
  printf("INFO: Done with benchmarking! Saving histogram to disk...\n");

  size_t written = fwrite(histogram, sizeof(uint32_t), HISTOGRAM_SIZE, output);

  if (written < HISTOGRAM_SIZE) {
    fprintf(stderr,
            "ERROR: Failed to write histogram to output file \"%s\": %s\n",
            output_path, strerror(errno));
    return EXIT_FAILURE;
  }

  printf("INFO: Done!\n");

  return EXIT_SUCCESS;
}
